{ ##
  @FILE                     UBDiff.pas
  @COMMENTS                 Contains main program logic.
  @PROJECT_NAME             BDiff
  @PROJECT_DESC             Outputs difference of binary files (Pascal
                            translation of original C code).
  @AUTHOR                   Original C program: Stefan Reuther <Streu@gmx.de>\
                            Pascal translation: Peter Johnson
                            <peter@delphidabbler.com>
  @WEBSITE                  Pascal versions: http://www.delphidabbler.com/
  @COPYRIGHT                Original C program: copyright  1999 by Stefan
                            Reuther.\
                            Pascal translation: copyright  2003 by Peter
                            Johnson.
  @LEGAL_NOTICE             Copying this program is allowed, as long as you
                            include original source code and Pascal translation
                            and document changes you made in a user-visible way
                            so people know they're using the new version, not
                            the original C code or Pascal translation.
  @HISTORY(
    @REVISION(
      @VERSION              1.0
      @DATE                 28/11/2003
      @COMMENTS             Original Pascal translation of original bdiff.c.
    )
  )
}


unit UBDiff;


{
  PASCAL IMPLEMENTATION NOTES:
    + The Pascal translation is only suitable for use on Windows targets - it
      compiles to a Win32 console application and uses the Windows API.
    + When shell redirection is used in the original C version running under
      MS-DOS/Windows to create binary difference files, the shell can write CRLF
      whenever LF appears in the output. This modification does not occur in the
      Pascal version and therefore it a safe to output binary files using
      redirection meaning it is not necessary to use the --output switch.
}


interface


{ the program's main interface code: called from the project file }
procedure Main;


implementation


{
  ABOUT THIS IMPLEMENTATION

  This version of the main bdiff program code is a literal translation of Stefan
  Reuther's bdiff.c file from BDiff v0.2.

  The Pascal translation is on the left in the following code, while the
  equivalent original C code is displayed on the right.
}


// ++++++++++++++++++++++++++++++++++++ // ++++++++++++++++++++++++++++++++++ //
// +        PASCAL TRANSLATION        + // +            ORIGINAL C          + //
// ++++++++++++++++++++++++++++++++++++ // ++++++++++++++++++++++++++++++++++ //
                                        //
                                        //
{$IOCHECKS OFF}                         //
                                        //
                                        //
uses                                    //
  // Delphi                             //
  SysUtils, Windows,                    //#include <stdio.h>
                                        //#include <stdlib.h>
                                        //#include <string.h>
                                        //#include <ctype.h>
  // Project: translation of blksort.c  //
  // C code lists imported fns here ->  //
  UBlkSort,                             //size_t* block_sort(char* data,
                                        //..size_t dlen);
                                        //size_t find_string(char* data,
                                        //..size_t* block, size_t len,
                                        //..char* sub, size_t max,
                                        //..size_t* index);
  // Project: special types & routines  //
  UBDiffTypes, UBDiffUtils;             //
                                        //
const                                   //
  VERSION_        = '0.2.1(p)'; {update}//#define VERSION "0.2"
  FORMAT_VERSION  = '02';               // #define FORMAT_VERSION "02"
  BUFFER_SIZE     = 4096;               // #define BUFFER_SIZE 4096
                                        //
                                        //
{ output format to use }                //
type TFormat = (FMT_BINARY,             // enum Format { FMT_BINARY,
  FMT_FILTERED, FMT_QUOTED);            //.. FMT_FILTERED, FMT_QUOTED };
                                        //
                                        //
{ structure for a matching block }      //
type                                    //
  TMatch = record                       //struct Match {
  oofs: size_t; {pos in old file}       //   size_t oofs;
  nofs: size_t; {pos in new file}       //   size_t nofs;
  len: size_t;  {length: 0 if no match} //   size_t len;
  end;                                  // };
  PMatch = ^TMatch;                     //
                                        //
{ for error messages }                  //
var progname: string;                   // char* progname;
                                        //
{ default options }                     //
var                                     //
  min_len: size_t = 24;                 // size_t min_len = 24;
  format: TFormat = FMT_QUOTED;         // enum Format format = FMT_QUOTED;
  verbose: Integer = 0;                 // int verbose = 0;
                                        //
{ format option handling }              //
type                                    //
  TFormatSpec = record                  // struct FormatSpec {
    header:                             //   void (*print_header)(char*, char*,
      procedure(oldfn, newfn: PChar;    //..   size_t, size_t);
        olds, news: size_t);            //
    add:                                //   void (*print_add)(char*, size_t);
      procedure(data: PShortInt;        //
        len: size_t);                   //
    copy:                               //   void (*print_copy)(char*, size_t,
      procedure(                        //..   char*, size_t, size_t);
        nbase: PShortIntArray;          //
        npos: size_t;                   //
        obase: PShortIntArray;          //
        opos: size_t; len: size_t);     //
  end;                                  // };
                                        //
procedure print_binary_header(oldfn,    // void print_binary_header(char*,
  newfn: PChar; oldl, newl: size_t);    //.. char*, size_t, size_t);
  forward;                              //
procedure print_text_header(oldfn,      // void print_text_header(char*, char*,
  newfn: PChar; olds, news: size_t);    //.. size_t, size_t);
  forward;                              //
procedure print_binary_add(             // void print_binary_add(char*, size_t);
  data: PShortInt; len: size_t);        //
  forward;                              //
procedure print_filtered_add(data:      // void print_filtered_add(char*,
  PShortInt; len: size_t); forward;     //.. size_t);
procedure print_quoted_add(             // void print_quoted_add(char*, size_t);
  data: PShortInt; len: size_t);        //
  forward;                              //
procedure print_text_copy(              // void print_text_copy(char*, size_t,
  nbase: PShortIntArray; npos: size_t;  //.. char*, size_t, size_t);
  obase: PShortIntArray; opos: size_t;  //
  len: size_t); forward;                //
procedure print_binary_copy(nbase:      // void print_binary_copy(char*, size_t,
  PShortIntArray; npos: size_t;         //.. char*, size_t, size_t);
  obase: PShortIntArray; opos: size_t;  //
  len: size_t); forward;                //
                                        //
var                                     //
  fmt_spec: array[TFormat] of           // struct FormatSpec fmt_spec[] = {
    TFormatSpec = (                     //
    ( header: print_binary_header;      //   { print_binary_header,
      add: print_binary_add;            //     print_binary_add,
      copy: print_binary_copy; ),       //     print_binary_copy },
    ( header: print_text_header;        //   { print_text_header,
      add: print_filtered_add;          //     print_filtered_add,
      copy: print_text_copy; ),         //     print_text_copy },
    ( header: print_text_header;        //   { print_text_header,
      add: print_quoted_add;            //     print_quoted_add,
      copy: print_text_copy; )          //     print_text_copy }
  );                                    // };
                                        //
{ exit program with error message }     //
procedure error_exit(msg: PChar);       //void error_exit(char* msg)
begin                                   //{
  WriteStrFmt(stderr, '%s: %s'#13#10,   //  fprintf(stderr, "%s: %s\n",
    [progname, msg]);                   //..  progname, msg);
  Halt(1);                              //  exit(1);
end;                                    //}
                                        //
{ load file, returning pointer to file  //
  data, exits with error message if out //
  of memory or not found }              //
function load_file(file_name: PChar;    //char* load_file(char* file_name,
  size_ret: Psize_t): PShortIntArray;   //..size_t* size_ret)
var                                     //
  fp: File of Char;                     //FILE* fp;
  data: PShortIntArray;                 //char* data;
  buffer: array[0..BUFFER_SIZE-1] of    //char buffer[BUFFER_SIZE];
    Char;                               //
  len: size_t;                          //size_t len;
  cur_len: size_t;                      //size_t cur_len;
  tmp: PShortIntArray;                  //char* tmp = realloc(data, cur_len
                                        //..+ len);
begin                                   //{
                                        //  FILE* fp;
                                        //  char* data;
                                        //  char buffer[BUFFER_SIZE];
                                        //  size_t len;
                                        //  size_t cur_len;
  { open file }                         //
  AssignFile(fp, file_name);            //  fp = fopen(file_name, "rb");
  Reset(fp);                            //
  if (IOResult <> 0) then               //  if(!fp) {
  begin                                 //
    perror(file_name);                  //
    Halt(1);                            //    exit(1);
  end;                                  //  }
                                        //
  { read file }                         //
  cur_len := 0;                         //  cur_len = 0;
  data := nil;                          //  data = 0;
  BlockRead(fp, buffer, BUFFER_SIZE,    //  while((len = fread(buffer, 1,
    len);                               //..  BUFFER_SIZE, fp))) {
  while (len > 0) do                    //
  begin                                 //
    tmp := data;                        //    char* tmp = realloc(data, cur_len
    ReallocMem(tmp, cur_len + len);     //..     + len);
    if not Assigned(tmp) then           //    if(!tmp) {
    begin                               //
      WriteStrFmt(stderr,               //      fprintf(stderr,
        '%s: Virtual memory exhausted'  //..      "%s: Virtual memory exhausted
        + #13#10, [progname]);          //..      \n", progname);
      Halt(1);                          //      exit(1);
    end;                                //    }
    data := tmp;                        //    data = tmp;
    Move(buffer, data[cur_len],         //    memcpy(data + cur_len, buffer,
       len);                            //..    len);
    Inc(cur_len, len);                  //    cur_len += len;
    BlockRead(fp, buffer, BUFFER_SIZE,  //  }
      len);                             //
  end;                                  //
  if not EOF(fp) then                   //  if(!feof(fp)) {
  begin                                 //
    perror(file_name);                  //    perror(file_name);
    CloseFile(fp);                      //    fclose(fp);
    Halt(1);                            //    exit(1);
  end;                                  //  }
                                        //
  { exit }                              //
  CloseFile(fp);                        //  fclose(fp);
  if Assigned(size_ret) then            //  if(size_ret)
    size_ret^ := cur_len;               //    *size_ret = cur_len;
  Result := data;                       //  return data;
end;                                    //}
                                        //
{ pack long in little-endian format     //
  into p }                              //
procedure pack_long(p: PShortInt;       //void pack_long(char* p, long l)
  l: Longint);                          //
begin                                   //{
  p^ := l and $FF;                      //  *p++ = l & 0xFF;
  Inc(p);                               //
  p^ := (l shr 8) and $FF;              //  *p++ = (l >> 8) & 0xFF;
  Inc(p);                               //
  p^ := (l shr 16) and $FF;             //  *p++ = (l >> 16) & 0xFF;
  Inc(p);                               //
  p^ := (l shr 24) and $FF;             //  *p   = (l >> 24) & 0xFF;
end;                                    //}
                                        //
{ compute simple checksum }             //
function checksum(data: PShortInt;      //long checksum(char* data, size_t len)
  len: size_t): Longint;                //
var                                     //
  l: Longint;                           //long l = 0;
begin                                   //{
  l := 0;                               //  long l = 0;
  while len <> 0 do                     //  while(len--) {
  begin                                 //
    Dec(len);                           //
    l := ((l shr 30) and 3)             //    l = ((l >> 30) & 3) | (l << 2);
      or (l shl 2);                     //
    l := l xor Ord(data^);              //    l ^= *data++;
    Inc(data);                          //
  end;                                  //  }
  Result := l;                          //  return l;
end;                                    //}
                                        //
{ print header for 'BINARY' format }    //
procedure print_binary_header(oldfn,    //void print_binary_header(char* oldfn,
  newfn: PChar; oldl, newl: size_t);    //.. char* newfn, size_t oldl, size_t
                                        //.. newl)
var                                     //
  head: array[0..15] of ShortInt;       //char head[16];
begin                                   //{
  Move('bdiff' + FORMAT_VERSION         //  strcpy(head, "bdiff" FORMAT_VERSION
    + #$1A, head[0], 8); {8 bytes}      //..  "\x1A");
  pack_long(@head[8], oldl);            //  pack_long(head + 8, oldl);
  pack_long(@head[12], newl);           //  pack_long(head + 12, newl);
  WriteBin(stdout, @head, 16);          //  fwrite(head, 1, 16, stdout);
end;                                    //}
                                        //
{ print header for text formats }       //
procedure print_text_header(oldfn,      //void print_text_header(char* oldfn,
  newfn: PChar; olds, news: size_t);    //.. char* newfn, size_t olds,
                                        //.. size_t news)
begin                                   //{
  WriteStrFmt(stdout,                   //  printf("%% --- %s (%d bytes)\n"
    '%% --- %s (%d bytes)'#13#10        //..  "%% +++ %s (%d bytes)\n",
    + '%% +++ %s (%d bytes)'#13#10,     //..  oldfn, olds, newfn, news);
    [oldfn, olds, newfn, news]          //
  );                                    //
end;                                    //}
                                        //
{ print data as C-escaped string }      //
procedure print_quoted_data(            //void print_quoted_data(char* data,
  data: PShortInt;                      //.. size_t len)
  len: size_t);                         //
begin                                   //{
  while (len <> 0) do                   //  while(len) {
  begin                                 //
    if isprint(Char(data^))             //    if(isprint(*data) && *data!='\\')
      and (Char(data^) <> '\') then     //
      WriteStr(stdout, Char(data^))     //      putchar(*data);
    else                                //    else
      WriteStr(stdout, '\'              //      printf("\\%03o", *data & 0xFF);
        + ByteToOct(data^ and $FF));    //
    Inc(data);                          //    data++;
    Dec(len);                           //    len--;
  end;                                  //  }
end;                                    //}
                                        //
{ print data with non-printing          //
  characters filtered }                 //
procedure print_filtered_data(          //void print_filtered_data(char* data,
  data: PShortInt; len: size_t);        //.. size_t len)
begin                                   //{
  while len <> 0  do                    //  while(len) {
  begin                                 //
    if isprint(Char(data^)) then        //    if(isprint(*data))
      WriteStr(stdout, Char(data^))     //      putchar(*data);
    else                                //    else
      WriteStr(stdout, '.');            //      putchar('.');
    Inc(data);                          //    data++;
    Dec(len);                           //    len--;
  end;                                  //  }
end;                                    //}
                                        //
{ print information for binary diff     //
  chunk }                               //
procedure print_binary_add(             //void print_binary_add(char* data,
  data: PShortInt; len: size_t);        //.. size_t len)
var                                     //
  buf: array[0..7] of ShortInt;         //char buf[8];
    { why 0..7? why not 0..3 - only 4   //
      bytes used }                      //
begin                                   //{
                                        //  char buf[8];
  WriteStr(stdout, '+');                //  putchar('+');
  pack_long(@buf[0], len);              //  pack_long(buf, len);
  WriteBin(stdout, @buf, 4);            //  fwrite(buf, 1, 4, stdout);
  WriteBin(stdout, data, len);          //  fwrite(data, 1, len, stdout);
end;                                    //}
                                        //
{ print information for filtered diff   //
  chunk }                               //
procedure print_filtered_add(           //void print_filtered_add(char* data,
  data: PShortInt; len: size_t);        //..  size_t len)
begin                                   //{
  WriteStr(stdout, '+');                //  putchar('+');
  print_filtered_data(data, len);       //  print_filtered_data(data, len);
  WriteStr(stdout, #13#10);             //  putchar('\n');
end;                                    //}
                                        //
{ print information for quoted diff     //
  chunk }                               //
procedure print_quoted_add(             //void print_quoted_add(char* data,
  data: PShortInt; len: size_t);        //.. size_t len)
begin                                   //{
  WriteStr(stdout, '+');                //  putchar('+');
  print_quoted_data(data, len);         //  print_quoted_data(data, len);
  WriteStr(stdout, #13#10);             //  putchar('\n');
end;                                    //}
                                        //
{ print information for copied data in  //
  text mode }                           //
procedure print_text_copy(              //void print_text_copy(char* nbase,
  nbase: PShortIntArray; npos: size_t;  //.. size_t npos, char* obase,
  obase: PShortIntArray; opos: size_t;  //.. size_t opos, size_t len)
  len: size_t);                         //
begin                                   //{
  WriteStrFmt(stdout,                   //  printf(
    '@ -[%d] => +[%d] %d bytes'         //..  "@ -[%d] => +[%d] %d bytes"\n ",
    + #13#10' ',                        //..  opos, npos, len);
    [opos, npos, len]                   //
  );                                    //
  if format = FMT_FILTERED then         //  if(format == FMT_FILTERED)
    print_filtered_data(@nbase[npos],   //    print_filtered_data(&nbase[npos],
      len)                              //..    len);
  else                                  //  else
    print_quoted_data(@nbase[npos],     //    print_quoted_data(&nbase[npos],
      len);                             //..    len);
  WriteStr(stdout, #13#10);             //  putchar('\n');
end;                                    //}
                                        //
{ print information for copied data in  //
  binary mode }                         //
procedure print_binary_copy(            //void print_binary_copy(char* nbase,
  nbase: PShortIntArray; npos: size_t;  //.. size_t npos, char* obase,
  obase: PShortIntArray; opos: size_t;  //.. size_t opos, size_t len)
  len: size_t);                         //
var                                     //
  rec: array[0..11] of ShortInt;        //char rec[12];
begin                                   //{
                                        //  char rec[12];
  WriteStr(stdout, '@');                //  putchar('@');
  pack_long(@rec[0], opos);             //  pack_long(rec + 0, opos);
  pack_long(@rec[4], len);              //  pack_long(rec + 4, len);
  pack_long(@rec[8],                    //  pack_long(rec + 8,
    checksum(@nbase[npos], len));       //..  checksum(&nbase[npos], len));
  WriteBin(stdout, @rec, 12);           //  fwrite(rec, 1, 12, stdout);
end;                                    //}
                                        //
{ find maximum-length match }           //
procedure bs_find_max_match(            //void bs_find_max_match(
  m_ret: PMatch;              { return }//..struct Match* m_ret,
  data: PShortIntArray; sort: PBlock;   //..char* data, size_t* sort,
  len: size_t;              { old file }//..size_t len,
  text: PShortInt; tlen: size_t);       //..char* text, size_t tlen)
                    { rest of new file }//..
var                                     //
  found_pos: size_t;                    //size_t found_pos;
  found_len: size_t;                    //size_t found_len = find_string(data,
                                        //..sort, len, text, tlen, &found_pos);
begin                                   //{
  m_ret^.len := 0;  {no match}          //  m_ret->len = 0;
  m_ret^.nofs := 0;                     //  m_ret->nofs = 0;
  while (tlen <> 0) do                  //  while(tlen) {
  begin                                 //
                                        //    size_t found_pos;
    found_len := find_string(data,      //    size_t found_len = find_string(
      sort, len, text, tlen,            //..    data, sort, len, text, tlen,
      @found_pos);                      //..    &found_pos);
    if found_len >= min_len then        //    if(found_len >= min_len) {
    begin                               //
      m_ret^.oofs := found_pos;         //      m_ret->oofs = found_pos;
      m_ret^.len := found_len;          //      m_ret->len = found_len;
      Exit;                             //      return;
    end;                                //    }
    Inc(text);                          //    text++;
    Inc(m_ret^.nofs);                   //    m_ret->nofs++;
    Dec(tlen);                          //    tlen--;
  end;                                  //  }
end;                                    //}
                                        //
{ print message, if enabled }           //
procedure log_status(const p: PChar);   //void log_status(const char* p)
begin                                   //{
  if verbose <> 0 then                  //  if(verbose)
    WriteStrFmt(stderr, '%s: %s'#13#10, //    fprintf(stderr, "%s: %s\n",
      [progname, p]);                   //..    progname, p);
end;                                    //}
                                        //
{ main routine: generate diff }         //
procedure bs_diff(fn, newfn: PChar);    //void bs_diff(char* fn, char* newfn)
var                                     //
  data: PShortIntArray;                 //char* data;
  data2: PShortIntArray;                //char* data2;
  len: size_t;                          //size_t len;
  len2, todo, nofs: size_t;             //size_t len2, todo, nofs;
  sort: PBlock;                         //size_t* sort;
  match: TMatch;                        //struct Match match;
begin                                   //{
                                        //  char* data;
                                        //  char* data2;
                                        //  size_t len;
                                        //  size_t len2, todo, nofs;
                                        //  size_t* sort;
  { initialize }                        //
  log_status('loading old file');       //  log_status("loading old file");
  data := load_file(fn, @len);          //  data = load_file(fn, &len);
  log_status('loading new file');       //  log_status("loading new file");
  data2 := load_file(newfn, @len2);     //  data2 = load_file(newfn, &len2);
  log_status('block sorting old file'); //  log_status(
                                        //..  "block sorting old file");
  sort := block_sort(data, len);        //  sort = block_sort(data, len);
  if not Assigned(sort) then            //  if(!sort) {
  begin                                 //
    WriteStrFmt(stderr,                 //    fprintf(stderr,
      '%s: virtual memory exhausted'    //..    "%s: virtual memory exhausted\n"
      + #13#10, [progname]);            //..    , progname);
    Halt(1);                            //    exit(1);
  end;                                  //  }
                                        //
  log_status('generating patch');       //  log_status("generating patch");
  fmt_spec[format].header(              //  fmt_spec[format].print_header(
    fn, newfn, len, len2);              //..  fn, newfn, len, len2);
                                        //
  { main loop }                         //
  todo := len2;                         //  todo = len2;
  nofs := 0;                            //  nofs = 0;
  while (todo <> 0) do                  //  while(todo) {
  begin                                 //
    { invariant: nofs + todo = len2 }   //
                                        //    struct Match match;
    bs_find_max_match(@match, data,     //    bs_find_max_match(&match, data,
      sort, len, @data2[nofs], todo);   //..    sort, len, &data2[nofs], todo);
    if match.len <> 0 then              //    if(match.len) {
    begin                               //
      { found a match }                 //
      if match.nofs <> 0 then           //      if(match.nofs != 0)
        { preceded by a "copy" block }  //
        fmt_spec[format].add(           //        fmt_spec[format].print_add(
          @data2[nofs], match.nofs);    //..        &data2[nofs], match.nofs);
      Inc(nofs, match.nofs);            //      nofs += match.nofs;
      Dec(todo, match.nofs);            //      todo -= match.nofs;
      fmt_spec[format].copy(data2, nofs,//      fmt_spec[format].print_copy(
        data, match.oofs, match.len);   //..      data2, nofs, data, match.oofs,
                                        //..      match.len);
      Inc(nofs, match.len);             //      nofs += match.len;
      Dec(todo, match.len);             //      todo -= match.len;
    end                                 //    } else {
    else                                //
    begin                               //
      fmt_spec[format].add(             //      fmt_spec[format].print_add(
        @data2[nofs], todo);            //..      &data2[nofs], todo);
      Break;                            //      break;
    end;                                //    }
  end;                                  //  }
  log_status('done');                   //  log_status("done");
end;                                    //}
                                        //
{ help & exit }                         //
procedure help;                         //void help()
begin                                   //{
  WriteStrFmt(stdout,                   //  printf("%s: binary `diff' -
    '%s: binary ''diff'' - compare two '//..  compare two binary files\n"
    + 'binary files'#13#10              //
    + #13#10                            //    "\n"
    + 'Usage: %s [options] old-file '   //    "Usage: %s [options] old-file
    + 'new-file [>patch-file]'#13#10    //..    new-file [>patch-file]\n"
    + #13#10                            //    "\n"
    + 'Valid options:'#13#10            //    "Valid options:\n"
    + ' -q                   Use '      //    " -q                   Use
    + 'QUOTED format'#13#10             //..    QUOTED format\n"
    + ' -f                   Use '      //    " -f                   Use
    + 'FILTERED format'#13#10           //..    FILTERED format\n"
    + ' -b                   Use '      //    " -b                   Use
    + 'BINARY format'#13#10             //..    BINARY format\n"
    + '       --format=FMT   Use'       //    "       --format=FMT   Use
    + 'specified format'#13#10          //..    specified format\n"
    + ' -m N  --min-equal=N  Minimum '  //    " -m N  --min-equal=N  Minimum
    + 'equal bytes to recognize an '    //..    equal bytes to recognize an
    + 'equal chunk'#13#10               //..    equal chunk\n"
    + ' -V    --verbose      Show '     //    " -V    --verbose      Show
    + 'status messages'#13#10           //..    status messages\n"
    + ' -h    --help         Show this '//    " -h    --help         Show this
    + 'help screen'#13#10               //..    help screen\n"
    + ' -v    --version      Show '     //    " -v    --version      Show
    + 'version information'#13#10       //..    version information\n"
    + #13#10                            //    "\n"
    + '(c) copyright 1999 Stefan '      //    "(c) copyright 1999 Stefan
    + 'Reuther <Streu@gmx.de>'#13#10,   //..    Reuther <Streu@gmx.de>\n",
    [progname, progname]);              //..    progname, progname);
  Halt(0);                              //  exit(0);
end;                                    //}
                                        //
{ version & exit }                      //
procedure version;                      //void version()
  function ExeDate: string;             //
  var                                   //
    H: Integer;                         //
    DOSDate: Integer;                   //
  begin                                 //
    H := FileOpen(ParamStr(0),          //
      fmOpenRead + fmShareDenyNone);    //
    try                                 //
      DOSDate := FileGetDate(H);        //
    finally                             //
      FileClose(H);                     //
    end;                                //
    Result := DateToStr(                //
      FileDateToDateTime(DOSDate));     //
  end;                                  //
begin                                   //{
  // NOTE: there is no Pascal equivalent//
  // of __DATE__ so we display update   //
  // date of program file rather than C //
  // version's compilation date         //
  WriteStrFmt(stdout, 'bdiff-%s %s '    //  printf("bdiff-" VERSION " ("
    + #13#10, [VERSION_, ExeDate]);     //..  __DATE__ ")\n");
  Halt(0);                              //  exit(0);
end;                                    //}
                                        //
{ read argument of --min-equal }        //
procedure set_min_equal(p: PChar);      //void set_min_equal(char* p)
var                                     //
  q: PChar;                             //char* q;
  x: LongWord;                          //unsigned long x;
begin                                   //{
                                        //  char* q;
                                        //  unsigned long x;
  if not Assigned(p) or (p^ = #0) then  //  if(!p || !*p)
    error_exit('Missing argument to '   //    error_exit("Missing argument to
      + '''--min-equal'' / ''-m''');    //..    `--min-equal' / `-m'");
  x := StrToULDec(p, q);                //  x = strtoul(p, &q, 0);
  if q^ <> #0 then                      //  if(*q)
    error_exit('Malformed number on '   //    error_exit("Malformed number on
      + 'command line');                //..    command line");
  if (x = 0) or (x > $7FFF) then        //  if(x == 0 || x > 0x7FFF)
    error_exit('Number out of range on '//    error_exit("Number out of range on
      + 'command line');                //..    command line");
  min_len := x;                         //  min_len = x;
end;                                    //}
                                        //
{ read argument of --format }           //
procedure set_format(p: PChar);         //void set_format(char* p)
begin                                   //{
  if not Assigned(p) then               //  if(!p)
    error_exit('Missing argument to '   //    error_exit("Missing argument to
      + '''--format''');                //..    `--format'");
  if StrComp(p, 'quoted') = 0 then      //  if(strcmp(p, "quoted")==0)
    format := FMT_QUOTED                //    format = FMT_QUOTED;
  else if (StrComp(p, 'filter') = 0)    //  else if(strcmp(p, "filter")==0 ||
    or (StrComp(p, 'filtered') = 0) then//..  strcmp(p, "filtered")==0)
    format := FMT_FILTERED              //    format = FMT_FILTERED;
  else if StrComp(p, 'binary') = 0 then //  else if(strcmp(p, "binary")==0)
    format := FMT_BINARY                //    format = FMT_BINARY;
  else                                  //  else
    error_exit(                         //    error_exit("Invalid format
      'Invalid format specification');  //..    specification");
end;                                    //}
                                        //
{ main routine: argument parsing }      //
                                        //
procedure Main;                         //int main(int argc, char** argv)
var                                     //
  oldfn: string;                        //char* oldfn = 0;
  newfn: string;                        //char* newfn = 0;
  outfn: string;                        //char* outfn = 0;
  i: Integer;                           //int i;
  p: PChar;                             //char* p = argv[i]+2
                                        //char* p = argv[i] + 1;
  fp: Integer;                          //FILE* p = freopen(outfn, "wb",stdout);
  argv: PChar;                          //
begin                                   //{
  oldfn := '';                          //  char* oldfn = 0;
  newfn := '';                          //  char* newfn = 0;
  outfn := '';                          //  char* outfn = 0;
                                        //  int i;
  progname := ExtractFileName(          //  progname = argv[0];
    ParamStr(0));                       //
                                        //
  i := 1;                               //  for(i = 1; i < argc; i++)
  while (i <= ParamCount) do            //
  begin                                 //
    argv := PChar(ParamStr(i) + #0#0#0);//
    if argv[0] = '-' then               //    if(argv[i][0] == '-') {
    begin                               //
      if argv[1] = '-' then             //      if(argv[i][1] == '-') {
      begin                             //
        { long option }                 //
        p := argv + 2;                  //        char* p = argv[i]+2;
        if StrComp(p, 'help') = 0 then  //        if(strcmp(p, "help")==0)
          help                          //          help();
        else if StrComp(p,              //        else if(strcmp(p,
          'version') = 0 then           //..        "version")==0)
          version                       //          version();
        else if StrComp(p,              //        else if(strcmp(p,
          'verbose') = 0 then           //..        "verbose")==0)
          verbose := 1                  //          verbose = 1;
        else if StrComp(p,              //        else if(strcmp(p,
          'output') = 0 then            //..        "output")==0)
        begin                           //
          Inc(i);                       //
          argv := PChar(ParamStr(i));   //
          if (argv^ = #0) then          //          if(!argv[++i]) {
          begin                         //
            WriteStrFmt(stderr,         //            fprintf(stderr, "%s:
              '%s: missing argument to '//..            missing argument to
              + '''--output'''#13#10,   //..            `--output'\n",
              [progname]);              //..            progname);
            Halt(1);                    //            return 1;
          end                           //          } else
          else                          //
            outfn := argv;              //            outfn=argv[i];
        end                             //
        else if StrLComp(p, 'output=',  //        else if(strncmp(p, "output=",
          7) = 0 then                   //..        7)==0)
          outfn := p + 7                //          outfn=p + 7;
        else if                         //        else if
          StrComp(p, 'format') = 0 then //..        (strcmp(p, "format")==0)
        begin                           //
          Inc(i);                       //
          argv := PChar(ParamStr(i));   //
          set_format(argv);             //          set_format(argv[++i]);
        end                             //
        else if StrLComp(p, 'format=',  //        else if(strncmp(p, "format=",
          7) = 0 then                   //..        7)==0)
          set_format(p + 7)             //          set_format(p + 7);
        else if StrComp(p,              //        else if(strcmp(p,
          'min-equal') = 0 then         //..        "min-equal")==0)
        begin                           //
          Inc(i);                       //
          argv := PChar(ParamStr(i));   //
          set_min_equal(argv);          //          set_min_equal(argv[++i]);
        end                             //
        else if StrLComp(p,             //        else if(strncmp(p,
          'min-equal=', 10) = 0 then    //          "min-equal=", 10)==0)
          set_min_equal(p + 10)         //          set_min_equal(p + 10);
        else                            //        else {
        begin                           //
          WriteStrFmt(stderr,           //          fprintf(stderr, "%s: unknown
            '%s: unknown option '       //..          option `--%s'\n"
            + '''--%s'''#13#10          //            "%s: try `%s --help' for
            + '%s: try ''%s --help'''   //..            more information\n",
            + 'for more information'    //
            + #13#10,                   //
            [progname, p, progname,     //            progname, p, progname,
            progname]);                 //..            progname);
          Halt(1);                      //          return 1;
        end;                            //        }
      end                               //      } else {
      else                              //
      begin                             //
        { short option }                //
        p := argv + 1;                  //        char* p = argv[i] + 1;
        while p^ <> #0 do               //        while(*p) {
        begin                           //
          case p^ of                    //          switch(*p) {
            'h':                        //            case 'h':
              help;                     //              help();
            'v':                        //            case 'v':
              version;                  //              version();
            'V':                        //            case 'V':
              verbose := 1;             //              verbose = 1;
                                        //              break;
            'q':                        //            case 'q':
              format := FMT_QUOTED;     //              format = FMT_QUOTED;
                                        //              break;
            'f':                        //            case 'f':
              format := FMT_FILTERED;   //              format = FMT_FILTERED;
                                        //              break;
            'b':                        //            case 'b':
              format := FMT_BINARY;     //              format = FMT_BINARY;
                                        //              break;
            'm':                        //            case 'm':
            begin                       //
              Inc(i);                   //
              argv :=                   //
                PChar(ParamStr(i));     //
              set_min_equal(argv);      //              set_min_equal(
                                        //..              argv[++i]);
                                        //              break;
            end;                        //
            'o':                        //            case 'o':
            begin                       //
              Inc(i);                   //
              argv :=                   //
                PChar(ParamStr(i));     //
              if argv^ = #0 then        //              if(!argv[++i]) {
              begin                     //
                WriteStrFmt(stderr,     //
                  '%s: missing argument'//                fprintf(stderr, "%s:
                  + ' to ''-o'''#13#10, //..                missing argument to
                  [progname]);          //..                `-o'\n", progname);
                Halt(1);                //                return 1;
              end                       //              } else
              else                      //
                outfn := argv;          //                outfn=argv[i];
                                        //              break;
            end;                        //
            else                        //            default:
              WriteStrFmt(stderr,       //              fprintf(stderr, "%s:
                '%s: unknown option '   //..              unknown option
                + '''-%s'''#13#10       //..              `-%c'\n"
                + '%s: try %s --help''' //..              "%s: try `%s --help'
                + ' for more'           //..              for more information
                + ' information'        //..              \n",
                + #13#10,               //..              progname, *p,
                [progname, p^, progname,//                progname, progname);
                progname]);             //
              Halt(1);                  //               return 1;
          end;                          //          }
          Inc(p);                       //          p++;
        end;                            //        }
      end;                              //      }
    end                                 //    } else {
    else                                //
    begin                               //
      if oldfn = '' then                //      if(!oldfn)
        oldfn := ParamStr(i)            //        oldfn = argv[i];
      else if newfn = '' then           //      else if(!newfn)
        newfn := ParamStr(i)            //        newfn = argv[i];
      else                              //      else
        error_exit('Too many file names'//        error_exit("Too many file
          + ' on command line');        //..        names on command line");
    end;                                //    }
    Inc(i);                             //
  end;                                  //
  if newfn = '' then                    //  if(!newfn)
    error_exit('Need two filenames');   //    error_exit("Need two filenames");
  if (outfn <> '')                      //  if(outfn &&
    and (outfn <> '-') then             //..  strcmp(outfn, "-") != 0) {
  begin                                 //
    fp := FileCreate(outfn);            //    FILE* p = freopen(outfn, "wb",
                                        //..    stdout);
    if fp <= 0 then                     //    if(!p) {
    begin                               //
      perror(outfn);                    //      perror(outfn);
      Halt(1);                          //      return 1;
    end;                                //    }
    Windows.SetStdHandle(               //
      STD_OUTPUT_HANDLE, fp);           //
  end;                                  //  }
                                        //
  bs_diff(PChar(oldfn), PChar(newfn));  //  bs_diff(oldfn, newfn);
                                        //  return 0;
end;                                    //}

end.
