{ ##
  @FILE                     UBDiffTypes.pas
  @COMMENTS                 Contains type declarations for Pascal version of
                            BDiff.
  @PROJECT_NAME             BDiff
  @PROJECT_DESC             Outputs difference of binary files (Pascal
                            translation of original C code).
  @AUTHOR                   Original C program: Stefan Reuther <Streu@gmx.de>\
                            Pascal translation: Peter Johnson
                            <peter@delphidabbler.com>
  @WEBSITE                  Pascal versions: http://www.delphidabbler.com/
  @COPYRIGHT                Original C program: copyright  1999 by Stefan
                            Reuther.\
                            Pascal translation: copyright  2003 by Peter
                            Johnson.
  @LEGAL_NOTICE             Copying this program is allowed, as long as you
                            include original source code and Pascal translation
                            and document changes you made in a user-visible way
                            so people know they're using the new version, not
                            the original C code or Pascal translation.
  @HISTORY(
    @REVISION(
      @VERSION              1.0
      @DATE                 28/11/2003
      @COMMENTS             Original version.
    )
  )
}


unit UBDiffTypes;


interface


type
  { size_t type is used extensively in C and in original code for this program.
    We therefore define the size_t and "pointer to size_t" types }
  size_t = Cardinal;
  Psize_t = ^size_t;

  { Some uses of *size_t in original C code actually reference an array of
    size_t and are referenced using array[] notation. The following types are
    declared to use in these circumstances to enabled similar notation in
    Pascal }
  TBlock = array[0..0] of size_t;
  PBlock = ^TBlock;

  { The original C code refers to the buffered file contents as an array of
    Char. The fact that Char is signed in C (-127..128) and unsigned in Pascal
    (0..255) means that the block sort algorithm and string lookup functions
    operate differently in C and Pascal. We therefore refer to the buffered file
    contents as an array of ShortInt in the Pascal code since ShortInt is
    defined as (-127..128). We declare a short int array type and pointer to it
    to enable us to asscess the buffer using array notation}
  TShortIntArray = array[0..(MaxInt div SizeOf(ShortInt) - 1)] of ShortInt;
  PShortIntArray = ^TShortIntArray;


implementation


end.
